#include <boost/asio.hpp>

#include "connection.hpp"
#include "ircmsg.hpp"
#include "linebuffer.hpp"
#include "settings.hpp"
#include "thread.hpp"
#include "write_irc.hpp"

#include "irc_parse_thread.hpp"
#include "registration_thread.hpp"
#include "ping_thread.hpp"
#include "watchdog_thread.hpp"
#include "snote_thread.hpp"

#include <algorithm>
#include <chrono>
#include <fstream>
#include <coroutine>
#include <iostream>
#include <limits>
#include <list>
#include <memory>
#include <string>
#include <string_view>
#include <tuple>
#include <utility>
#include <variant>
#include <vector>
#include <unordered_map>
#include <unordered_set>

using namespace std::chrono_literals;

auto unhandled_message_thread(Connection * connection) -> void
{
    connection->add_listener<IrcMsgEvent>([](IrcMsgEvent const& event)
    {
        if (IrcCommand::UNKNOWN == event.command)
        {
            auto& irc = event.irc;
            std::cout << "Unknown message " << irc.command;
            for (auto const arg : irc.args)
            {
                std::cout << " " << arg;
            }
            std::cout << "\n";
        }
    });
}

auto start(boost::asio::io_context & io, Settings const& settings) -> void
{
    auto connection = std::make_shared<Connection>(io);

    watchdog_thread(connection.get());
    irc_parse_thread(connection.get());
    ping_thread(connection.get());
    registration_thread(connection.get(), settings.password, settings.username, settings.realname, settings.nickname);
    snote_thread(connection.get());
    unhandled_message_thread(connection.get());

    boost::asio::co_spawn(
        io,
        connection->connect(io, settings.host, settings.service),
        [&io, &settings](std::exception_ptr e)
        {
            auto timer = std::make_shared<boost::asio::steady_timer>(io);
            timer->expires_from_now(5s);
            timer->async_wait([&io, &settings, timer](auto) {
                start(io, settings);
            });
        });
}

auto get_settings() -> Settings
{
    if (auto config_stream = std::ifstream {"config.toml"})
    {
        return Settings::from_stream(config_stream);
    }
    else
    {
        std::cerr << "Unable to open config.toml\n";
        std::exit(1);
    }
}

auto main() -> int
{
    auto const settings = get_settings();
    auto io = boost::asio::io_context{};
    start(io, settings);
    io.run();
}
